class CDFparser
{
public:
	enum state_types 
	{	st_begin,
		st_record_open, st_record_member, st_record_close,
		st_list_open, st_list_close, 
		st_ident,
		st_int, 
		st_in_string, st_string_end,
		st_end, st_error
	};
	state_types state;
	char string_char;
	bool as_hex;
	long int_val;
	char name[100];
	lword linenr;
	lword columnnr;
	char *error_text;
	
	CDFparser() : state(st_end), text_acc(0) {}
	~CDFparser() { close(); }
	
	void open(AbstractTextAccessor *new_text_acc)
	{	text_acc = new_text_acc;
		state = st_begin;
		more = new_text_acc->more;
		ch = new_text_acc->ch;
		depth = 0;
		linenr = 1;
		columnnr = 1;
		next();
	}
	inline void close()
	{	text_acc = 0;
		state = st_end;
	}
	void next()
	{
		if (state >= st_end)
			return;
		if (!more)
		{	close();
			return;
		}
				
		if (state == st_in_string)
		{	next_string_char();
			return;
		}	

		while (more && isspace(ch))
			next_char();
		
		if (!more)
		{	if (depth > 0) 
			{	state = st_error;
				error_text = "Closing brackets missing";
			}
			else
				state = st_end;
			return;
		}
		
		if (ch == ')')
		{
			if (depth == 0 || ss[depth-1] != ')')
			{	state = st_error;
				error_text = "Unexpected ')'";
				return;
			}
			depth--;
			state = st_record_close;
			next_char();
			return;
		}
		if (ch == ']')
		{
			if (depth == 0 || ss[depth-1] != ']')
			{	state = st_error;
				error_text = "Unexpected ']'";
				return;
			}
			depth--;
			state = st_list_close;
			next_char();
			return;
		}
		
		if (ch == '{')
		{	state = st_in_string;
			next_char();
			next_string_char();
			return;
		}
		if (isdigit(ch))
		{	state = st_int;
			int_val = 0;
			do
			{ 	int_val = 10*int_val + ch - '0';
				next_char();
			}
			while (more && isdigit(ch));
			return;
		}			
		
		*name = '\0';
		if (isalpha(ch))
		{
			int i = 0;
			while (isalnum(ch))
			{	if (i < 99)
					name[i++] = ch;
				next_char();
			}
			name[i] = '\0';
		}
		
		if (!more && *name != '\0')
		{	state = st_ident;
			return;
		}
		if (ch == ':')
		{	if (depth == 0 || ss[depth-1] != ')')
			{	state = st_error;
				error_text = "Can not use ':' here";
				return;
			}
			state = st_record_member;
		}	
		else if (ch == '(')
		{	state = st_record_open;
			if (depth == 1000)
			{	state = st_error;
				error_text = "Too many open brackets";
				return;
			} 
			ss[depth++] = ')';
			next_char();
		}
		else if (ch == '[')
		{	state = st_list_open;
			if (depth == 1000)
			{	state = st_error;
				error_text = "Too many open brackets";
				return;
			} 
			ss[depth++] = ']';
			next_char();
		}
		else if (*name != '\0')
		{	state = st_ident;
			return;
		}
		else
		{	state = st_error;
			error_text = "I am confused";
		}
		return;
	}

	inline int current_depth()
	{	return depth;
	}
	void skip_till_depth(int requested_depth)
	{
		while (state < st_end && depth > requested_depth)
			next();
		if (state == st_record_close || state == st_list_close)
			next();
		else
		{	state = st_error;
			error_text = "Missing close brackets at end-of-file";
		}
	}
	void skip_till_close()
	{	skip_till_depth(depth-1);
	}
	
	void next_string_char()
	{
		as_hex = false;
		if (state >= st_end)
			return;
			
		if (state != st_in_string)
		{	state = st_error;
			error_text = "Not in string";
			return;
		}

		while (more && ch == '\n')
		{	next_char();
			while (more && (ch == ' ' || ch == '\t'))
				next_char();
		}
		if (!more)
		{	state = st_error;
			error_text = "Reaching end-of-file while in string";
			return;
		}
		if (ch == '}')
		{	state = st_string_end;
			next_char();
		}
		else if (ch == '#')
		{	next_char();
			if (!more)
			{	state = st_error;
				error_text = "Expecting some character after #";
				return;
			}
			else if (isdigit(ch))
				string_char = (ch - '0') << 4;
			else if ('A' <= ch && ch <= 'F')
				string_char = (ch - 'A' + 10) << 4;
			else if ('a' <= ch && ch <= 'f')
				string_char = (ch - 'a' + 10) << 4;
			else
			{	state = st_error;
				error_text = "Rubbish after #";
				return;
			}
			next_char();
			if (!more)
			{	state = st_error;
				error_text = "Expecting two characters after #";
				return;
			}
			else if (isdigit(ch))
				string_char |= (ch - '0');
			else if ('A' <= ch && ch <= 'F')
				string_char |= (ch - 'A' + 10);
			else if ('a' <= ch && ch <= 'f')
				string_char |= (ch - 'a' + 10);
			else
			{	state = st_error;
				error_text = "Rubbish after #";
				return;
			}
			next_char();
			as_hex = true;
		}
		else
		{	string_char = ch;
			next_char();
		}
	}			
	
private:
	AbstractTextAccessor *text_acc;
	bool more;
	char ch;

	word depth;
	char ss[1000];
	
	inline void next_char()
	{	if (ch == '\n')
		{	linenr++;
			columnnr = 1;
		}
		else
			columnnr++;
		text_acc->next();
		more = text_acc->more;
		ch = text_acc->ch;
		while (more && ch == '%' && columnnr == 1)
		{
			while (more && ch != '\n')
			{	columnnr++;
				text_acc->next();
				more = text_acc->more;
				ch = text_acc->ch;
			}
			if (more && ch == '\n')
			{	linenr++;
				columnnr = 1;
				text_acc->next();
				more = text_acc->more;
				ch = text_acc->ch;
			}
		}
	}

};

