#include "stdafx.h"
#include "CppSyntaxColorer.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

CppSyntaxColorState::CppSyntaxColorState()
{
	state = 0;
}

CppSyntaxColorState::CppSyntaxColorState(CppSyntaxColorState* original)
{
	stateStack = original->stateStack;
	state = original->state;
}

//	C++ keywords (MSVC5.0 + POET5.0)
static LPTSTR s_apszCppKeywordList[] =
{
	_T("__asm"),
	_T("enum"),
	_T("__multiple_inheritance"),
	_T("template"),
	_T("auto"),
	_T("__except"),
	_T("__single_inheritance"),
	_T("this"),
	_T("__based"),
	_T("explicit"),
	_T("__virtual_inheritance"),
	_T("thread"),
	_T("bool"),
	_T("extern"),
	_T("mutable"),
	_T("throw"),
	_T("break"),
	_T("false"),
	_T("naked"),
	_T("true"),
	_T("case"),
	_T("__fastcall"),
	_T("namespace"),
	_T("try"),
	_T("catch"),
	_T("__finally"),
	_T("new"),
	_T("__try"),
	_T("__cdecl"),
	_T("float"),
	_T("operator"),
	_T("typedef"),
	_T("char"),
	_T("for"),
	_T("private"),
	_T("typeid"),
	_T("class"),
	_T("friend"),
	_T("protected"),
	_T("typename"),
	_T("const"),
	_T("goto"),
	_T("public"),
	_T("union"),
	_T("const_cast"),
	_T("if"),
	_T("register"),
	_T("unsigned"),
	_T("continue"),
	_T("inline"),
	_T("reinterpret_cast"),
	_T("using"),
	_T("__declspec"),
	_T("__inline"),
	_T("return"),
	_T("uuid"),
	_T("default"),
	_T("int"),
	_T("short"),
	_T("__uuidof"),
	_T("delete"),
	_T("__int8"),
	_T("signed"),
	_T("virtual"),
	_T("dllexport"),
	_T("__int16"),
	_T("sizeof"),
	_T("void"),
	_T("dllimport"),
	_T("__int32"),
	_T("static"),
	_T("volatile"),
	_T("do"),
	_T("__int64"),
	_T("static_cast"),
	_T("wmain"),
	_T("double"),
	_T("__leave"),
	_T("__stdcall"),
	_T("while"),
	_T("dynamic_cast"),
	_T("long"),
	_T("struct"),
	_T("xalloc"),
	_T("else"),
	_T("main"),
	_T("switch"),
	_T("interface"),
	//	Added by a.s.
	_T("persistent"),
	_T("_persistent"),
	_T("transient"),
	_T("depend"),
	_T("ondemand"),
	_T("transient"),
	_T("cset"),
	_T("useindex"),
	_T("indexdef"),
	NULL
};

static BOOL IsCppKeyword(LPCTSTR pszChars, int nLength)
{
	for (int L = 0; s_apszCppKeywordList[L] != NULL; L ++)
	{
		if (strncmp(s_apszCppKeywordList[L], pszChars, nLength) == 0
				&& s_apszCppKeywordList[L][nLength] == 0)
			return TRUE;
	}
	return FALSE;
}

static BOOL IsCppNumber(LPCTSTR pszChars, int nLength)
{
	if (nLength > 2 && pszChars[0] == '0' && pszChars[1] == 'x')
	{
		for (int I = 2; I < nLength; I++)
		{
			if (isdigit(pszChars[I]) || (pszChars[I] >= 'A' && pszChars[I] <= 'F') ||
										(pszChars[I] >= 'a' && pszChars[I] <= 'f'))
				continue;
			return FALSE;
		}
		return TRUE;
	}
	if (! isdigit(pszChars[0]))
		return FALSE;
	for (int I = 1; I < nLength; I++)
	{
		if (! isdigit(pszChars[I]) && pszChars[I] != '+' &&
			pszChars[I] != '-' && pszChars[I] != '.' && pszChars[I] != 'e' &&
			pszChars[I] != 'E')
			return FALSE;
	}
	return TRUE;
}

#define COOKIE_COMMENT			0x0001
#define COOKIE_PREPROCESSOR		0x0002
#define COOKIE_EXT_COMMENT		0x0004
#define COOKIE_STRING			0x0008
#define COOKIE_CHAR				0x0010

#define COLOR_WHITESPACE	RGB(255, 255, 255)
#define COLOR_NORMALTEXT	RGB(  0,   0,   0)
#define COLOR_PREPROCESSOR	RGB(128,   0,   0)
#define COLOR_COMMENT		RGB( 64,  64, 220)
#define COLOR_NUMBER		RGB(128,   0,   0)
#define COLOR_OPERATOR		RGB(  0,   0,   0)
#define COLOR_STRING		RGB(  0, 128,   0)
#define COLOR_BK_STRING		RGB(240, 255, 240)
#define COLOR_KEYWORD		RGB(  0,  64, 128)
#define COLOR_SELBKGND		RGB(  0,   0,   0)
#define COLOR_SELTEXT		RGB(255, 255, 255)
#define COLOR_TEXT_ERROR	RGB(128, 0, 0)
#define COLOR_BK_ERROR		RGB(255, 240, 240)

CppSyntaxColorer::CppSyntaxColorer()
{
	defaultTextColor = ::GetSysColor(COLOR_WINDOWTEXT);
	defaultBkColor = ::GetSysColor(COLOR_WINDOW);
	noTextColor = defaultBkColor;
	selectTextColor = RGB(255, 255, 255);
	selectBkColor = RGB(0, 0, 0);
}

void CppSyntaxColorer::parseLine(const char* line, CppSyntaxColorState &Cppstate, int &lineFlags)
{
	SharedStack<char> *stateStack = &Cppstate.stateStack;
	int state = Cppstate.state;

	BOOL bFirstChar     = (state & ~COOKIE_EXT_COMMENT) == 0;
	BOOL bRedefineBlock = TRUE;
	BOOL bDecIndex  = FALSE;
	int nIdentBegin = -1;
	for (int I = 0; ; I++)
	{
		if (bRedefineBlock)
		{
			int nPos = I;
			if (bDecIndex)
				nPos--;
			if (state & (COOKIE_COMMENT | COOKIE_EXT_COMMENT))
			{
				startColor(nPos, COLOR_COMMENT, COLOR_WHITESPACE);
			}
			else
			if (state & (COOKIE_CHAR | COOKIE_STRING))
			{
				startColor(nPos, COLOR_STRING, COLOR_BK_STRING);
			}
			else
			if (state & COOKIE_PREPROCESSOR)
			{
				startColor(nPos, COLOR_PREPROCESSOR, COLOR_WHITESPACE);
			}
			else
			{
				startColor(nPos, COLOR_NORMALTEXT, COLOR_WHITESPACE);
			}
			bRedefineBlock = FALSE;
			bDecIndex      = FALSE;
		}

		if (line[I] == '\0')
			break;

		if (state & COOKIE_COMMENT)
		{
			startColor(I, COLOR_COMMENT, COLOR_WHITESPACE);
			state |= COOKIE_COMMENT;
			break;
		}

		//	String constant "...."
		if (state & COOKIE_STRING)
		{
			if (line[I] == '"' && (I == 0 || line[I - 1] != '\\' || (I >= 2 && line[I - 2] == '\\')))
			{
				state &= ~COOKIE_STRING;
				bRedefineBlock = TRUE;
				endColor(I);
			}
			continue;
		}

		//	Char constant '..'
		if (state & COOKIE_CHAR)
		{
			if (line[I] == '\'' && (I == 0 || line[I - 1] != '\\' || (I >= 2 && line[I - 2] == '\\')))
			{
				state &= ~COOKIE_CHAR;
				bRedefineBlock = TRUE;
				endColor(I);
			}
			continue;
		}

		//	Extended comment /*....*/
		if (state & COOKIE_EXT_COMMENT)
		{
			if (I > 0 && line[I] == '/' && line[I - 1] == '*')
			{
				state &= ~COOKIE_EXT_COMMENT;
				bRedefineBlock = TRUE;
			}
			continue;
		}

		if (I > 0 && line[I] == '/' && line[I - 1] == '/')
		{
			startColor(I - 1, COLOR_COMMENT, COLOR_WHITESPACE);
			state |= COOKIE_COMMENT;
			break;
		}

		//	Preprocessor directive #....
		if (state & COOKIE_PREPROCESSOR)
		{
			if (I > 0 && line[I] == '*' && line[I - 1] == '/')
			{
				startColor(I - 1, COLOR_COMMENT, COLOR_WHITESPACE);
				state |= COOKIE_EXT_COMMENT;
			}
			continue;
		}

		//	Normal text
		if (line[I] == '"')
		{
			startColor(I+1, COLOR_STRING, COLOR_BK_STRING);
			state |= COOKIE_STRING;
			continue;
		}
		if (line[I] == '\'')
		{
			startColor(I+1, COLOR_STRING, COLOR_BK_STRING);
			state |= COOKIE_CHAR;
			continue;
		}
		if (I > 0 && line[I] == '*' && line[I - 1] == '/')
		{
			startColor(I - 1, COLOR_COMMENT, COLOR_WHITESPACE);
			state |= COOKIE_EXT_COMMENT;
			continue;
		}

		if (bFirstChar)
		{
			if (line[I] == '#')
			{
				startColor(I, COLOR_PREPROCESSOR, COLOR_WHITESPACE);
				state |= COOKIE_PREPROCESSOR;
				continue;
			}
			if (! isspace(line[I]))
				bFirstChar = FALSE;
		}


		if (nIdentBegin >= 0 && !(isalnum(line[I]) || line[I] == '_' || line[I] == '.'))
		{
			if (IsCppKeyword(line + nIdentBegin, I - nIdentBegin))
			{
				startColor(nIdentBegin, COLOR_KEYWORD, COLOR_WHITESPACE);
			}
			else if (IsCppNumber(line + nIdentBegin, I - nIdentBegin))
			{
				startColor(nIdentBegin, COLOR_NUMBER, COLOR_WHITESPACE);
			}
			bRedefineBlock = TRUE;
			bDecIndex = TRUE;
			nIdentBegin = -1;
		}

		if (line[I] == '(')
		{	
			stateStack->push(')');
		}
		else if (line[I] == ')')
		{	
			if (stateStack->topIs(')'))
				stateStack->pop();
			else
				startColor(I, COLOR_TEXT_ERROR, COLOR_BK_ERROR);
		}
		else if (line[I] == '[')
		{	
			stateStack->push(']');
		}
		else if (line[I] == ']')
		{	if (stateStack->topIs(']'))
				stateStack->pop();
			else
				startColor(I, COLOR_TEXT_ERROR, COLOR_BK_ERROR);
		}
		else if (line[I] == '{')
		{	
			if (stateStack->topIs(')') || stateStack->topIs(']'))
				startColor(I, COLOR_TEXT_ERROR, COLOR_BK_ERROR);
			stateStack->push('}');
		}
		else if (line[I] == '}')
		{	
			if (stateStack->topIs('}'))
				stateStack->pop();
			else
				startColor(I, COLOR_TEXT_ERROR, COLOR_BK_ERROR);
		}
		else if (isalnum(line[I]) || line[I] == '_' || line[I] == '.')
		{
			if (nIdentBegin == -1)
				nIdentBegin = I;
		}
		else
		{
		}
	}

	if (nIdentBegin >= 0)
	{
		if (IsCppKeyword(line + nIdentBegin, I - nIdentBegin))
		{
			startColor(nIdentBegin, COLOR_KEYWORD, COLOR_WHITESPACE);
		}
		else
		if (IsCppNumber(line + nIdentBegin, I - nIdentBegin))
		{
			startColor(nIdentBegin, COLOR_NUMBER, COLOR_WHITESPACE);
		}
	}

	if (I > 0 && line[I - 1] != '\\')
		state &= COOKIE_EXT_COMMENT;

	Cppstate.state = state;
}
