#include "stdafx.h"
#include "SampleView.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

// ---- Greek characters -------------

#define ALPHA		1
#define BETA		2
#define GAMMA		3
#define DELTA		4
#define EPSILON		5
#define ZETA		6
#define ETA			7
#define THETA		8
#define IOTA		9
#define KAPPA		10
#define LAMBDA		11	
#define MU			12
#define NU			13
#define XI			14
#define OMICRON		15
#define PI			16
#define RHO			17
#define SIGMA		18
#define TAU			19
#define UPSILON		20
#define PHI			21
#define CHI			22
#define PSI			23
#define OMEGA		24
#define LOW_I_PRE	25
#define CAP      (1<<6)
#define ACUTE    (1<<7)
#define GRAVE    (1<<8)
#define UML		 (1<<9)
#define TILDE    (1<<10)
#define ASP      (1<<11)
#define NASP     (1<<12)

#define GREEKCHAR(X)   ((X)&(31))
#define GREEKCAP(X)    ((X)&(CAP))
#define GREEKCHARCAP(X) ((X)&(31|CAP))
#define GREEKSTRESS(X) ((X)&(ACUTE|GRAVE))
#define GREEKTILDE(X)    ((X)&(TILDE))
#define GREEKACC(X)    ((X)&(UML|TILDE))
#define GREEKASP(X)    ((X)&(ASP|NASP))
#define GREEKCHARSTRESS(X) ((X)&(31|ACUTE|GRAVE))


#define NR_GREEK_CHAR  152
struct 
{	char ch;
	long gcc;
} greek_encoding[NR_GREEK_CHAR] = {
	{ 0x27,			 NASP							},
	{ 0x41,	ALPHA				|CAP				},
	{ 0x42,	BETA				|CAP				},
	{ 0x44,	IOTA	|NASP	|GRAVE					},
	{ 0x45,	EPSILON				|CAP				},
	{ 0x46,								LOW_I_PRE	},
	{ 0x47,	UPSILON	|NASP	|ACUTE					},
	{ 0x48,	ETA					|CAP				},
	{ 0x49,	IOTA				|CAP				},
	{ 0x4A,	OMEGA	|NASP	|GRAVE					},
	{ 0x4B,	KAPPA				|CAP				},
	{ 0x4C,	ALPHA	|ASP	|GRAVE					},
	{ 0x4D,	MU					|CAP				},
	{ 0x4E,	NU					|CAP				},
	{ 0x4F,	OMICRON				|CAP				},
	{ 0x50,	RHO					|CAP				},
	{ 0x51,	IOTA	|ASP	|GRAVE					},
	{ 0x52,	UPSILON	|ASP	|GRAVE					},
	{ 0x53,	OMEGA	|ASP	|GRAVE					},
	{ 0x54,	TAU					|CAP				},
	{ 0x55,	IOTA				|UML|GRAVE			},
	{ 0x56,	UPSILON				|UML|GRAVE			},
	{ 0x58,	CHI					|CAP				},
	{ 0x59,	UPSILON				|CAP				},
	{ 0x5A,	ZETA				|CAP				},
	{ 0x5C,			 NASP							},
	{ 0x5E,			 ASP							},
	{ 0x61,	ALPHA			|GRAVE					},
	{ 0x62,	EPSILON			|GRAVE					},
	{ 0x63,	ETA				|GRAVE					},
	{ 0x64,	IOTA			|GRAVE					},
	{ 0x65,	OMICRON			|GRAVE					},
	{ 0x66,	UPSILON			|GRAVE					},
	{ 0x67,	OMEGA			|GRAVE					},
	{ 0x69,	ALPHA	|NASP	|GRAVE					},
	{ 0x6A,	ETA		|NASP	|GRAVE					},
	{ 0x6B,	OMICRON	|NASP	|GRAVE					},
	{ 0x6D,	ETA		|ASP	|GRAVE					},
	{ 0x6E,	OMICRON	|ASP	|GRAVE					},
	{ 0x6F,	OMICRON	|ASP	|GRAVE					},
	{ 0x70,	ALPHA	|NASP	|TILDE					},
	{ 0x71,	ETA		|NASP	|TILDE					},
	{ 0x72,	IOTA	|NASP	|TILDE					},
	{ 0x73,	UPSILON	|NASP	|TILDE					},
	{ 0x74,	OMEGA	|NASP	|TILDE					},
	{ 0x77,	ETA		|ASP	|TILDE					},
	{ 0x78,	IOTA	|ASP	|TILDE					},
	{ 0x79,	UPSILON	|ASP	|TILDE					},
	{ 0x7A,	OMEGA	|ASP	|TILDE					},
	{ 0x80,	ALPHA			|TILDE					},
	{ 0x83,	ETA				|TILDE					},
	{ 0x84,	ETA										},
	{ 0x85,	IOTA			|TILDE					},
	{ 0x86,	UPSILON			|TILDE					},
	{ 0x87,	OMEGA			|TILDE					},
	{ 0x88,	ALPHA	|NASP							},
	{ 0x89,	EPSILON	|NASP							},
	{ 0x8A,	ETA		|NASP							},
	{ 0x8B,					 ACUTE					},
	{ 0x8C,	IOTA	|NASP							},
	{ 0x8D,	OMICRON	|NASP							},
	{ 0x8E,	UPSILON	|NASP							},
	{ 0x8F,	OMEGA	|NASP							},
	{ 0x90,	ALPHA	|ASP							},
	{ 0x91,	EPSILON	|ASP							},
	{ 0x92,	EPSILON	|ASP	|GRAVE					},
	{ 0x93,	ETA		|ASP							},
	{ 0x94,	IOTA	|ASP							},
	{ 0x95,	OMICRON	|ASP							},
	{ 0x96,	UPSILON	|ASP							},
	{ 0x97,	OMEGA	|ASP							},
	{ 0x98,			 NASP	|ACUTE					},
	{ 0x99,	ALPHA	|NASP	|ACUTE					},
	{ 0x9A,	EPSILON	|NASP	|ACUTE					},
	{ 0x9B,	ETA		|NASP	|ACUTE					},
	{ 0x9C,	IOTA	|NASP	|ACUTE					},
	{ 0x9D,	OMICRON	|NASP	|ACUTE					},
	{ 0x9E,	UPSILON	|NASP	|ACUTE					},
	{ 0x9F,	OMEGA	|NASP	|ACUTE					},
	{ 0xA1,	GAMMA				|CAP				},
	{ 0xA2,	DELTA				|CAP				},
	{ 0xA3,	THETA				|CAP				},
	{ 0xA4,	LAMBDA				|CAP				},
	{ 0xA5,	CHI					|CAP				},
	{ 0xA6,	PI					|CAP				},
	{ 0xA8,	RHO		|ASP							},
	{ 0xA9,	THETA									},
	{ 0xAA,	SIGMA				|CAP				},
	{ 0xAD,			 ASP	|ACUTE					},
	{ 0xB0,	ALPHA				|CAP				},
	{ 0xB1,	ALPHA	|ASP	|ACUTE					},
	{ 0xB2,	EPSILON	|ASP	|ACUTE					},
	{ 0xB3,	ETA		|ASP	|ACUTE					},
	{ 0xB4,	IOTA	|ASP	|ACUTE					},
	{ 0xB5,	BETA				|CAP				},
	{ 0xB6,	EPSILON				|CAP				},
	{ 0xB7,	ZETA				|CAP				},
	{ 0xB8,	ETA					|CAP				},
	{ 0xB9,	IOTA				|CAP				},
	{ 0xBA,	KAPPA				|CAP				},
	{ 0xBB,	MU					|CAP				},
	{ 0xBC,	PHI					|CAP				},
	{ 0xBE, PSI					|CAP				},
	{ 0xBF,	OMEGA				|CAP				},
	{ 0xC0,	ALPHA			|ACUTE					},
	{ 0xC1,	NU					|CAP				},
	{ 0xC2,	OMICRON	|ASP	|GRAVE					},
	{ 0xC3, OMICRON				|CAP				},
	{ 0xC4,	PI					|CAP				},
	{ 0xC5,	UPSILON	|ASP	|ACUTE					},
	{ 0xC6,	TAU					|CAP				},
	{ 0xCC,	CHI					|CAP				},
	{ 0xCF,	OMEGA	|ASP	|ACUTE					},
	{ 0xD5,			 NASP							},
	{ 0xD6,	ALPHA									},
	{ 0xD8,	IOTA	|NASP		|CAP				},
	{ 0xDB,	EPSILON			|ACUTE					},
	{ 0xDC,	ETA				|ACUTE					},
	{ 0xDD,	IOTA			|ACUTE					},
	{ 0xDE,	OMICRON			|ACUTE					},
	{ 0xE0,	UPSILON			|ACUTE					},
	{ 0xE1,	ALPHA									},
	{ 0xE2,	BETA									},
	{ 0xE3,	PSI										},
	{ 0xE4,	DELTA									},
	{ 0xE5,	EPSILON									},
	{ 0xE6,	PHI										},
	{ 0xE7,	GAMMA									},
	{ 0xE8,	ETA										},
	{ 0xE9,	IOTA									},
	{ 0xEA,	XI										},
	{ 0xEB,	KAPPA									},
	{ 0xEC,	LAMBDA									},
	{ 0xED,	MU										},
	{ 0xEE,	NU										},
	{ 0xEF,	OMICRON									},
	{ 0xF0,	PI										},
	{ 0xF1,	OMEGA			|GRAVE					},
	{ 0xF2,	RHO										},
	{ 0xF3,	SIGMA									},
	{ 0xF4,	TAU										},
	{ 0xF5,	THETA									},
	{ 0xF6,	OMEGA									},
	{ 0xF7,	SIGMA									},
	{ 0xF8,	CHI										},
	{ 0xF9,	UPSILON									},
	{ 0xFA,	ZETA									},
	{ 0xFB,	IOTA				|UML				},
	{ 0xFC,	UPSILON				|UML				},
	{ 0xFD,	IOTA				|UML|ACUTE			},
	{ 0xFE,	UPSILON				|UML|ACUTE			},
	{ 0xFF,	OMEGA				|TILDE				},
};

bool valid_greek_char_code(long gcc)
{
	for (int i = 1; i < NR_GREEK_CHAR; i++)
		if (greek_encoding[i].gcc == gcc)
			return true;

	return false;
}

bool correct_greek(char *greek_trans, int &error_pos)
{	
	char *start = greek_trans;
	for (; *greek_trans != '\0'; )
	{
		if (*greek_trans == ' ')
			greek_trans++;
		else if (*greek_trans == '1')
			greek_trans++;
		else if (*greek_trans == '2')
			greek_trans++;
		else if (*greek_trans == '3')
			greek_trans++;
		else if (   *greek_trans == ',' || *greek_trans == '.' 
				 || *greek_trans == '[' || *greek_trans == ']')
			greek_trans++;
		else if (*greek_trans == ';')
			greek_trans++;
		else if (*greek_trans == '#')
		{	if (isxdigit(greek_trans[1]) && isxdigit(greek_trans[2]))
			{	greek_trans += 3;
			}				
			else
				greek_trans++;
		}
		else
		{
			long gcc = 0;
	
			bool go = true;
			for (; *greek_trans != '\0' && go; greek_trans++)
			{
				go = false;
				switch(*greek_trans)
				{	case '~'	: gcc |= TILDE; 		go = true; break;
					case '\''	: gcc |= NASP; 			go = true; break;
					case 'h'	: gcc |= ASP; 			go = true; break;
					case '"'	: gcc |= UML;	 		go = true; break;
					case '\\'	: gcc |= GRAVE; 		go = true; break;
					case '/'	: gcc |= ACUTE; 		go = true; break;
					case 'A'	: gcc |= ALPHA|CAP; 	break;
					case 'B'	: gcc |= BETA|CAP; 		break;
					case 'C'	: if (greek_trans[1] == 'h') { 
					              gcc |= CHI|CAP; 		greek_trans++; break; }
								  break;
					case 'D'	: gcc |= DELTA|CAP; 	break;
					case 'E'	: if (greek_trans[1] == '^') {
								  gcc |= ETA|CAP;  		greek_trans++; break; }
								  gcc |= EPSILON|CAP;	break;
					case 'G'	: gcc |= GAMMA|CAP; 	break;
					case 'I'	: gcc |= IOTA|CAP; 		break;
					case 'K'	: gcc |= KAPPA|CAP;		break;
					case 'L'	: gcc |= LAMBDA|CAP;	break;
					case 'M'	: gcc |= MU|CAP;		break;
					case 'N'	: gcc |= NU|CAP;		break;
					case 'O'	: if (greek_trans[1] == '^') {
								  gcc |= OMEGA|CAP;		greek_trans++; break; }
								  gcc |= OMICRON|CAP;	break;
					case 'P'	: if (greek_trans[1] == 'h') {
								  gcc |= PHI|CAP;		greek_trans++; break; }
								  if (greek_trans[1] == 's') {
								  gcc |= PSI|CAP;		greek_trans++; break; }
								  gcc |= PI|CAP;		break;
					case 'S'	: gcc |= SIGMA|CAP;		break;
					case 'R'	: gcc |= RHO|CAP;		break;
					case 'T'	: if (greek_trans[1] == 'h') {
								  gcc |= THETA|CAP;		greek_trans++; break; }
								  gcc |= TAU|CAP;		break;
					case 'X'	: gcc |= XI|CAP;		break;
					case 'Y'	: gcc |= UPSILON|CAP;	break;
					case 'Z'	: gcc |= ZETA|CAP; 		break;
					case 'a'	: gcc |= ALPHA;			break;
					case 'b'	: gcc |= BETA;			break;
					case 'c'	: if (greek_trans[1] == 'h') { 
					              gcc |= CHI; 			greek_trans++; break; }
								  break;
					case 'd'	: gcc |= DELTA;			break;
					case 'e'	: if (greek_trans[1] == '^') {
								  gcc |= ETA;			greek_trans++; break; }
								  gcc |= EPSILON;		break;
					case 'g'	: gcc |= GAMMA;			break;
					case 'i'	: if (greek_trans[1] == '_') {
								  gcc |= LOW_I_PRE;		greek_trans++; break; }
								  gcc |= IOTA;			break;
					case 'k'	: gcc |= KAPPA;			break;
					case 'l'	: gcc |= LAMBDA;		break;
					case 'm'	: gcc |= MU;			break;
					case 'n'	: gcc |= NU;			break;
					case 'o'	: if (greek_trans[1] == '^') {
								  gcc |= OMEGA;			greek_trans++; break; }
								  gcc |= OMICRON;		break;
					case 'p'	: if (greek_trans[1] == 'h') {
								  gcc |= PHI;			greek_trans++; break; }
								  if (greek_trans[1] == 's') {
								  gcc |= PSI;			greek_trans++; break; }
								  gcc |= PI;			break;
					case 'r'	: gcc |= RHO;			break;
					case 's'	: gcc |= SIGMA;			break;
					case 't'	: if (greek_trans[1] == 'h') {
								  gcc |= THETA;			greek_trans++; break; }
								  gcc |= TAU;			break;
					case 'u'	: gcc |= UPSILON;		break;
					case 'x'	: gcc |= XI;			break;
					case 'z'	: gcc |= ZETA;			break;
					default: go = false; greek_trans--; break;
				}
			}
			if (   valid_greek_char_code(gcc)
				|| (   valid_greek_char_code(gcc & ~(31|CAP))
					&& valid_greek_char_code(GREEKCHARCAP(gcc))))
				; // okay
			else
			{	error_pos = greek_trans - start;
				return false;
			}
		}
	}
	return true;
}


#define DEFINE_BLOCK(pos, colorindex)	\
	ASSERT((pos) >= 0 && (pos) <= nLength);\
	if (pBuf != NULL)\
	{\
		if (nActualItems == 0 || pBuf[nActualItems - 1].m_nCharPos <= (pos)){\
		pBuf[nActualItems].m_nCharPos = (pos);\
		pBuf[nActualItems].m_nColorIndex = (colorindex);\
		nActualItems ++;}\
	}

#define STATE(X) 	(X & 31)
#define ST_TOP		 0
#define ST_ERROR	 1
#define ST_GEOM		 2
#define ST_PAR		 3
#define ST_TEXT		 4
#define ST_BOLD		 5
#define ST_BOLD_G	 6
#define ST_BOLD_H	 7
#define ST_IT		 8
#define ST_IT_P		 9
#define ST_IT_G		10
#define ST_G		11
#define ST_H		12
#define ST_P		13
#define ST_P_G		14
#define ST_P_H		15
#define ST_END		16

#define ST_IN_TEXT	32


#define SYM_UNDEFINED		0
#define SYM_LIST_OPEN		1
#define SYM_LIST_CLOSE		2
#define SYM_REC_OPEN		3
#define SYM_REC_CLOSE		4
#define SYM_IDENT			5
#define SYM_LIST_NAME_OPEN	6
#define SYM_REC_NAME_OPEN	7
#define SYM_NUMBER			8


DWORD CSampleView::ParseLine(DWORD dwParseState, int nLineIndex, TEXTBLOCK *pBuf, int &nActualItems)
{
	int nLength = GetLineLength(nLineIndex);
	if (nLength <= 0)
		return dwParseState;

	if (dwParseState == ST_ERROR)
	{	DEFINE_BLOCK(0, COLORINDEX_PREPROCESSOR);
		return ST_ERROR;
	}

	LPCTSTR pszChars    = GetLineChars(nLineIndex);
	char buf[100000];

	if (nLength >= 2 && pszChars[0] == '%' || pszChars[1] == '%')
	{
		DEFINE_BLOCK(0, COLORINDEX_COMMENT);
		return dwParseState;
	}

	int i = 0;
	if (pszChars[0] == '\t' || pszChars[0] == ' ')
	{
		while (i < nLength && (pszChars[i] == '\t' || pszChars[i] == ' '))
			i++;
		DEFINE_BLOCK(0, COLORINDEX_NORMALTEXT);
	}

	if (dwParseState & ST_IN_TEXT)
	{
		int k = 0;
		int begin = i;
		for (; i < nLength && pszChars[i] != '}'; i++)
			if (k < 99999)
				buf[k++] = pszChars[i];
		buf[k] = '\0';
		if (pszChars[i] == '}')
		{	dwParseState &= ~ST_IN_TEXT;
			i++;
		}
		if (STATE(dwParseState) == ST_G)
		{	int error_pos;
			DEFINE_BLOCK(begin, COLORINDEX_KEYWORD);
			if (!correct_greek(buf, error_pos))
			{	DEFINE_BLOCK(begin + error_pos + 1, COLORINDEX_PREPROCESSOR);
				return ST_ERROR;
			}
		}
		else
		{	DEFINE_BLOCK(begin, COLORINDEX_STRING);
		}
	}

	while (i < nLength)
	{
		// scan symbol
		int dwNewParseState = ST_ERROR;
		int color = COLORINDEX_NORMALTEXT;

		int begin = i;
		if (pszChars[i] == '[')
		{	i++;
			
			switch (dwParseState)
			{
			case ST_TOP:	dwNewParseState = ST_GEOM; break;
			}
		}
		else if (pszChars[i] == ']')
		{	i++;
			
			switch (dwParseState)
			{
			case ST_GEOM:	dwNewParseState = ST_END; break;
			case ST_PAR:	dwNewParseState = ST_GEOM; break;
			case ST_TEXT:	dwNewParseState = ST_PAR; break;
			case ST_BOLD:	dwNewParseState = ST_TEXT; break;
			case ST_BOLD_G:	dwNewParseState = ST_BOLD; break;
			case ST_BOLD_H:	dwNewParseState = ST_BOLD; break;
			case ST_IT:		dwNewParseState = ST_TEXT; break;
			case ST_IT_P:	dwNewParseState = ST_IT; break;
			case ST_IT_G:	dwNewParseState = ST_IT; break;
			case ST_G:		dwNewParseState = ST_TEXT; break;
			case ST_H:		dwNewParseState = ST_TEXT; break;
			case ST_P:		dwNewParseState = ST_TEXT; break;
			case ST_P_G:	dwNewParseState = ST_P; break;
			case ST_P_H:	dwNewParseState = ST_P; break;
			}
		}
		else if (pszChars[i] == '(')
		{	i++;

			// should not occur
		}
		else if (pszChars[i] == ')')
		{	i++;

			// should not occur
		}
		else if (isalpha(pszChars[i]))
		{	int k = 0;
			for (; i < nLength && isalnum(pszChars[i]); i++)
				if (k < 99999)
					buf[k++] = pszChars[i];
			buf[k] = '\0';
			if (pszChars[i] == '[')
			{	i++;
				switch (dwParseState)
				{
				case ST_GEOM:
					if (   strcmp(buf, "norm") == 0
						|| strcmp(buf, "col1") == 0
						|| strcmp(buf, "col2") == 0
						|| strcmp(buf, "col3") == 0
						|| strcmp(buf, "col4") == 0
						|| strcmp(buf, "vn1") == 0
						|| strcmp(buf, "vn2") == 0
						|| strcmp(buf, "vn3") == 0
						|| strcmp(buf, "vn4") == 0
						|| strcmp(buf, "peri") == 0
						|| strcmp(buf, "other") == 0)
						dwNewParseState = ST_PAR;
					break;
				case ST_PAR:
					if (   strcmp(buf, "par") == 0
						|| strcmp(buf, "number") == 0
						|| strcmp(buf, "grieks") == 0
						|| strcmp(buf, "trans") == 0
						|| strcmp(buf, "vertaling") == 0
						|| strcmp(buf, "a1") == 0
						|| strcmp(buf, "a2") == 0
						|| strcmp(buf, "a3") == 0
						|| strcmp(buf, "kopwoord") == 0)
						dwNewParseState = ST_TEXT;
					break;
				case ST_TEXT:
					if (strcmp(buf, "B") == 0)
						dwNewParseState = ST_BOLD;
					else if (strcmp(buf, "I") == 0)
						dwNewParseState = ST_IT;
					else if (strcmp(buf, "G") == 0)
						dwNewParseState = ST_G;
					else if (strcmp(buf, "H") == 0)
						dwNewParseState = ST_H;
					else if (strcmp(buf, "P") == 0)
						dwNewParseState = ST_P;
					break;
				case ST_BOLD:
					if (strcmp(buf, "G") == 0)
						dwNewParseState = ST_BOLD_G;
					else if (strcmp(buf, "H") == 0)
						dwNewParseState = ST_BOLD_H;
				case ST_IT:
					if (strcmp(buf, "P") == 0)
						dwNewParseState = ST_IT_P;
					else if (strcmp(buf, "G") == 0)
						dwNewParseState = ST_IT_G;
					break;
				case ST_P:
					if (strcmp(buf, "G") == 0)
						dwNewParseState = ST_P_G;
					else if (strcmp(buf, "H") == 0)
						dwNewParseState = ST_P_H;
					break;
				}
			}
			else if (pszChars[i] == '(')
			{	i++;

				switch (dwParseState)
				{
				case ST_TEXT:
				case ST_BOLD:
				case ST_IT:
				case ST_IT_P:
				case ST_H:
				case ST_P:
					if (strcmp(buf, "t") == 0)
					{	if (i < nLength && pszChars[i] == ')')
						{	i++;
							dwNewParseState = dwParseState;
						}
					}
					else if (   strcmp(buf, "S") == 0
							 || strcmp(buf, "S1") == 0
							 || strcmp(buf, "S2") == 0)
					{	
						if (isdigit(pszChars[i]))
						{	for (i++; i < nLength && isdigit(pszChars[i]); i++);
							
							if (i < nLength && pszChars[i] == ')')
							{	i++;
								dwNewParseState = dwParseState;
							}
						}
					}
				}
			}
			else
			{
				// not valid
			}
		}
		else if (pszChars[i] == '{')
		{
			switch (dwParseState)
			{
			case ST_TEXT:
			case ST_BOLD:
			case ST_BOLD_G:
			case ST_BOLD_H:
			case ST_IT:
			case ST_IT_P:
			case ST_IT_G:
			case ST_G:
			case ST_H:
			case ST_P:
			case ST_P_G:
			case ST_P_H:

				i++;
				dwNewParseState = dwParseState | ST_IN_TEXT;
				
				int k = 0;
				for (; i < nLength && pszChars[i] != '}'; i++)
					if (k < 99999)
						buf[k++] = pszChars[i];
				buf[k] = '\0';
				if (pszChars[i] == '}')
				{	dwNewParseState &= ~ST_IN_TEXT;
					i++;
				}
				if (   STATE(dwParseState) == ST_G
					|| STATE(dwParseState) == ST_P_G
					|| STATE(dwParseState) == ST_IT_G
					|| STATE(dwParseState) == ST_BOLD_G)
				{	
					color = COLORINDEX_KEYWORD;
					int error_pos;
					if (!correct_greek(buf, error_pos))
					{	DEFINE_BLOCK(begin, color);
						DEFINE_BLOCK(begin + error_pos + 1, COLORINDEX_PREPROCESSOR);
						return ST_ERROR;
					}
				}
				else
				{	color = COLORINDEX_STRING;
				}
			}
		}
		if (dwNewParseState == ST_ERROR)
			color = COLORINDEX_PREPROCESSOR;

		DEFINE_BLOCK(begin, color);
		dwParseState = dwNewParseState;
		if (dwNewParseState == ST_ERROR)
			break;
	}

	return dwParseState;
}
